METHOD_CONNECTION_START = 10
METHOD_CONNECTION_START_OK = 11
METHOD_CONNECTION_SECURE = 20
METHOD_CONNECTION_SECURE_OK = 21
METHOD_CONNECTION_TUNE = 30
METHOD_CONNECTION_TUNE_OK = 31
METHOD_CONNECTION_OPEN = 40
METHOD_CONNECTION_OPEN_OK = 41
METHOD_CONNECTION_CLOSE = 50
METHOD_CONNECTION_CLOSE_OK = 51

METHOD_CHANNEL_OPEN = 10
METHOD_CHANNEL_OPEN_OK = 11
METHOD_CHANNEL_FLOW = 20
METHOD_CHANNEL_FLOW_OK = 21
METHOD_CHANNEL_CLOSE = 40
METHOD_CHANNEL_CLOSE_OK = 41

METHOD_EXCHANGE_DECLARE = 10
METHOD_EXCHANGE_DECLARE_OK = 11
METHOD_EXCHANGE_DELETE = 20
METHOD_EXCHANGE_DELETE_OK = 21
METHOD_EXCHANGE_BIND = 30  # rabbitmq extension
METHOD_EXCHANGE_BIND_OK = 31  # rabbitmq extension

METHOD_QUEUE_DECLARE = 10
METHOD_QUEUE_DECLARE_OK = 11
METHOD_QUEUE_BIND = 20
METHOD_QUEUE_BIND_OK = 21
METHOD_QUEUE_PURGE = 30
METHOD_QUEUE_PURGE_OK = 31
METHOD_QUEUE_DELETE = 40
METHOD_QUEUE_DELETE_OK = 41
METHOD_QUEUE_UNBIND = 50
METHOD_QUEUE_UNBIND_OK = 51

METHOD_BASIC_QOS = 10
METHOD_BASIC_QOS_OK = 11
METHOD_BASIC_CONSUME = 20
METHOD_BASIC_CONSUME_OK = 21
METHOD_BASIC_CANCEL = 30
METHOD_BASIC_CANCEL_OK = 31
METHOD_BASIC_PUBLISH = 40
METHOD_BASIC_RETURN_ = 50
METHOD_BASIC_DELIVER = 60
METHOD_BASIC_GET = 70
METHOD_BASIC_GET_OK = 71
METHOD_BASIC_GET_EMPTY = 72
METHOD_BASIC_ACK = 80
METHOD_BASIC_REJECT = 90
METHOD_BASIC_RECOVER_ASYNC = 100
METHOD_BASIC_RECOVER = 110
METHOD_BASIC_RECOVER_OK = 111
METHOD_BASIC_NACK = 120  # RabbitMQ extension

# whole class is a RabbitMQ extension
METHOD_CONFIRM_SELECT = 10
METHOD_CONFIRM_SELECT_OK = 11
